"""
Wrapper around torch.distributions.transforms to allow for joint transforms on TensorDicts.
"""
import weakref
from typing import Any, Generic, Optional, TypeVar, Union

import torch
from torch import nn
from torch.distributions.transforms import Transform, _InverseTransform

T_co = TypeVar("T_co", covariant=True, bound="TypedTransform")


class _TransformRef(Generic[T_co]):
    """A covariant reference to a transform.

    Particularily used to allow subclasses to refine the inverse type.
    """

    weak_ref: Optional[weakref.ReferenceType[T_co]]

    def __init__(self, obj: T_co):
        self.weak_ref = weakref.ref(obj)

    def __call__(self) -> Optional[T_co]:
        return self.weak_ref() if self.weak_ref is not None else None


X = TypeVar("X", covariant=False)
Y = TypeVar("Y", covariant=False)


class TypedTransform(Generic[X, Y], Transform):
    """Transforms with typed argument and return values.

    Should not be instantiated directly, but rather through the `TypedTransform` class.

    Notes:
        Special care needs to be taken in overriding the inverse. In particular, the weak reference is of a
        non-covariant generic type, and therefore isn't compatible with subclasses returning refined inverses. We solve
        this by using a custom covariant reference class.
    """

    def __call__(self, x: X) -> Y:  # pylint: disable=useless-parent-delegation
        return super().__call__(x)  # type: ignore

    def _inv_call(self, y: Y) -> X:  # pylint: disable=useless-parent-delegation
        return super()._inv_call(y)  # type: ignore

    def _call(self, x: X) -> Y:  # pylint: disable=useless-parent-delegation
        return super()._call(x)

    def _inverse(self, y: Y) -> X:  # pylint: disable=useless-parent-delegation
        return super()._inverse(y)

    def log_abs_det_jacobian(self, x: X, y: Y):  # pylint: disable=useless-parent-delegation
        return super().log_abs_det_jacobian(x, y)

    @property
    def inv(self) -> "TypedTransform[Y, X]":
        inv: Optional[TypedTransform[Y, X]] = None
        match self._inv:
            case _TransformRef():
                inv = self._inv()
            case TypedTransform():
                inv = self._inv
        if inv is None:
            inv = _TypedInverseTransform[Y, X](self)
            self._inv: Union[_TransformRef[_TypedInverseTransform[Y, X]], TypedTransform[Y, X]] = _TransformRef[
                _TypedInverseTransform[Y, X]
            ](inv)
        return inv


class _TypedInverseTransform(Generic[Y, X], TypedTransform[Y, X], _InverseTransform):
    _inv: TypedTransform[X, Y]


class TransformModule(Generic[X, Y], TypedTransform[X, Y], nn.Module):
    """Transforms with learnable parameters.

    This is similar to the `pyro.distributions.torch_transform.TransformModule` class.
    """

    def __hash__(self):
        """Return the nn.Module based hash.

        Notes:
            The Transformation hash is None.
        """
        return super(torch.nn.Module, self).__hash__()

    @property
    def inv(self) -> "TransformModule[Y, X]":
        inv: Optional[TransformModule[Y, X]] = None
        match self._inv:
            case _TransformRef():
                inv = self._inv()
            case TransformModule():
                inv = self._inv
        if inv is None:
            inv = _InverseTransformModule[Y, X](self)
            self._inv: Union[_TransformRef[_InverseTransformModule[Y, X]], TransformModule[Y, X]] = _TransformRef[
                _InverseTransformModule[Y, X]
            ](inv)
        return inv


class _InverseTransformModule(Generic[Y, X], TransformModule[Y, X], _TypedInverseTransform):
    """Inverse transformation of a TransformModule."""

    _inv: TransformModule[X, Y]


class SequentialTransformModule(Generic[X, Y], TransformModule[X, Y], nn.Sequential):
    """Sequential transform with TransformModule transformations."""

    def __init__(self, *args: TransformModule[X, Y], cache_size: int = 0):
        super().__init__(cache_size=cache_size)
        for idx, module in enumerate(args):
            self.add_module(str(idx), module)

    def _inverse(self, y: Y) -> X:
        current: Any = y
        # Create a tuple of submodules before reversing, to avoid a copy of the Sequential module
        for module in tuple(self)[::-1]:
            assert isinstance(module, TransformModule)
            current = module.inv(current)
        return current

    def _call(self, x: X) -> Y:
        return nn.Sequential.__call__(self, x)



"""Module that provides data normalization functionality."""
from typing import Callable, Optional

import torch
import torch.distributions as td
from tensordict import TensorDictBase
from torch import nn

from causica.distributions.transforms import JointTransformModule, SequentialTransformModule, TransformModule

Normalizer = TransformModule[TensorDictBase, TensorDictBase]
FitNormalizerType = Callable[[TensorDictBase], Normalizer]


class LoadNoneTensorMixin(nn.Module):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self._register_load_state_dict_pre_hook(self._update_tensor_size_on_load)

    def _update_tensor_size_on_load(self, state_dict: dict[str, torch.Tensor], prefix: str, *args, **kwargs) -> None:
        _ = args, kwargs
        for key, value in state_dict.items():
            local_key = key.removeprefix(prefix)
            if hasattr(self, local_key) and getattr(self, local_key) is None:
                setattr(self, local_key, torch.empty_like(value))


class Standardizer(TransformModule[torch.Tensor, torch.Tensor], td.AffineTransform, LoadNoneTensorMixin):
    """Standardizer module for a single variable, ie a single tensor."""

    def __init__(self, mean: Optional[torch.Tensor], std: Optional[torch.Tensor], *args, **kwargs) -> None:
        """
        Args:
            mean: Mean of the variable
            std: Standard deviation of the variable
            *args, **kwargs: Passed to the AffineTransform
        """
        loc = scale = None
        if mean is not None and std is not None:
            loc = -mean / std
            scale = 1 / std
        super().__init__(loc, scale, *args, **kwargs)
        del self.loc, self.scale  # Unset these temporarily to allow registering as buffers

        self.loc: torch.Tensor
        self.scale: torch.Tensor
        self.register_buffer("loc", loc)
        self.register_buffer("scale", scale)


def fit_standardizer(data: TensorDictBase) -> Normalizer:
    """Return a standardizer that updates data to zero mean and unit standard deviation."""
    means = data.apply(
        lambda x: torch.mean(x, dim=0, keepdim=False),
        batch_size=torch.Size(),
    )
    # Filter out std == 0
    stds = data.apply(
        lambda x: torch.std(x, dim=0, keepdim=False),
        batch_size=torch.Size(),
    ).apply(lambda x: torch.where(x == 0, torch.ones_like(x), x))

    return JointTransformModule({key: Standardizer(means.get(key), stds.get(key)) for key in means.keys()})


def chain_normalizers(*fit_functions: FitNormalizerType) -> FitNormalizerType:
    """Chain a number of normalizers together.

    Args:
        *fit_functions: Functions that produce normalizers.

    Returns:
        A function that fits the sequence of normalizers.
    """

    def sequential_fitting(X: TensorDictBase) -> Normalizer:
        transform_modules = []
        for fit_function in fit_functions:
            transform_module = fit_function(X)
            X = transform_module(X)
            transform_modules.append(transform_module)
        return SequentialTransformModule[TensorDictBase, TensorDictBase](*transform_modules)

    return sequential_fitting



import abc
from typing import Mapping

import torch
import torch.distributions as td

from causica.data_generation.samplers.sampler import Sampler
from causica.distributions import JointNoiseModule
from causica.distributions.noise import NoiseModule, UnivariateNormalNoiseModule
from causica.distributions.noise.bernoulli import BernoulliNoiseModule


class NoiseModuleSampler(Sampler[NoiseModule]):
    """
    An interface of a univariate noise sampler
    """

    @abc.abstractmethod
    def sample(
        self,
    ) -> NoiseModule:
        """Sample a sample type with given shape"""


class JointNoiseModuleSampler(NoiseModuleSampler):
    """Sampler for JointNoiseModule, given shapes and types of different variables"""

    def __init__(
        self,
        noise_dist_samplers: Mapping[str, NoiseModuleSampler],
    ):
        super().__init__()
        self.noise_dist_samplers = noise_dist_samplers

    def sample(self) -> JointNoiseModule:
        noise_modules = {}
        for key, noise_sampler in self.noise_dist_samplers.items():
            noise_modules[key] = noise_sampler.sample()
        return JointNoiseModule(independent_noise_modules=noise_modules)


class UnivariateNormalNoiseModuleSampler(NoiseModuleSampler):
    """Sample a UnivariateNormalNoiseModule, with standard deviation given by a distribution."""

    def __init__(self, std_dist: td.Distribution, dim: int = 1):
        super().__init__()
        self.std_dist = std_dist
        self.dim = dim

    def sample(
        self,
    ):
        return UnivariateNormalNoiseModule(dim=self.dim, init_log_scale=torch.log(self.std_dist.sample()).item())


class BernoulliNoiseModuleSampler(NoiseModuleSampler):
    """Sample a BernoulliNoiseModule, with base_logits given by a distribution."""

    def __init__(self, base_logits_dist: td.Distribution, dim: int = 1):
        super().__init__()
        self.base_logits_dist = base_logits_dist
        self.dim = dim

    def sample(
        self,
    ) -> NoiseModule:
        base_logits = self.base_logits_dist.sample().item()
        return BernoulliNoiseModule(dim=self.dim, init_base_logits=base_logits)



from __future__ import annotations

import torch
import torch.distributions as td

from causica.data_generation.samplers.sampler import Sampler
from causica.functional_relationships.functional_relationships import FunctionalRelationships
from causica.functional_relationships.linear_functional_relationships import LinearFunctionalRelationships
from causica.functional_relationships.rff_functional_relationships import RFFFunctionalRelationships


class FunctionalRelationshipsSampler(Sampler[FunctionalRelationships]):
    """Abstract class for sampling Functional Relationships."""

    def __init__(self, shapes_dict: dict[str, torch.Size]) -> None:
        super().__init__()

        self.shapes_dict = shapes_dict


class LinearRelationshipsSampler(FunctionalRelationshipsSampler):
    """Sample a Linear Functional Relationship, by providing a distribution for the coefficient matrix."""

    def __init__(self, scale_dist: td.Distribution, shapes_dict: dict[str, torch.Size]):
        super().__init__(shapes_dict)
        self.scale_dist = scale_dist

    def sample(self) -> FunctionalRelationships:
        return LinearFunctionalRelationships(
            shapes=self.shapes_dict, initial_linear_coefficient_matrix=self.scale_dist.sample()
        )


class RFFFunctionalRelationshipsSampler(FunctionalRelationshipsSampler):
    """Sample a Non Linear Functional Relationship, by providing two distributions:
    a first distribution for the random features, and a second distribution for the linear outer coefficients."""

    def __init__(self, rf_dist: td.Distribution, coeff_dist: td.Distribution, shapes_dict: dict[str, torch.Size]):
        super().__init__(shapes_dict)
        self.rf_dist = rf_dist
        self.coeff_dist = coeff_dist

    def sample(self) -> FunctionalRelationships:
        return RFFFunctionalRelationships(
            shapes=self.shapes_dict,
            initial_random_features=self.rf_dist.sample(),
            initial_coefficients=self.coeff_dist.sample(),
        )



from itertools import product
from typing import Iterable

import pytorch_lightning as pl
import torch
import torch.distributions as td
from tensordict import TensorDict
from torch.utils.data import ChainDataset, DataLoader, Dataset

from causica.data_generation.samplers.functional_relationships_sampler import (
    LinearRelationshipsSampler,
    RFFFunctionalRelationshipsSampler,
)
from causica.data_generation.samplers.noise_dist_sampler import (
    JointNoiseModuleSampler,
    UnivariateNormalNoiseModuleSampler,
)
from causica.data_generation.samplers.sampler import Sampler
from causica.data_generation.samplers.sem_sampler import SEMSampler
from causica.data_generation.synthetic_dataset import CausalDataset
from causica.distributions import ErdosRenyiDAGDistribution
from causica.functional_relationships.functional_relationships import FunctionalRelationships


class SyntheticDataModule(pl.LightningDataModule):
    """A datamodule to produce datasets and their underlying causal graphs and interventions."""

    def __init__(
        self,
        node_nums: list[int],
        probs: list[float],
        train_batch_size: int,
        test_batch_size: int,
        dataset_size: int,
        num_interventions: int = 0,
        num_sems: int = 0,
        num_workers: int = 0,
    ) -> None:
        """
        Args:
            node_nums: List of number of nodes for generated graphs
            probs: List of the probabilities of edges in each graph when using Erdos Renyi
            train_batch_size: The training batch size to use
            test_batch_size: The testing batch size to use
            dataset_size: The size of dataset to generate
            num_interventions: The number of interventions to generate (0 for no interventions)
            num_sems: The number of SEMs to generate (0 for infinite SEMs)
            num_workers: The number of workers to use for the dataloader
        """
        super().__init__()
        self.node_nums = node_nums
        self.probs = probs
        self.dataset_size = dataset_size
        self.train_batch_size = train_batch_size
        self.test_batch_size = test_batch_size
        self.sems_dict: dict[tuple[int, float, bool], SEMSampler] = {}
        self.num_interventions = num_interventions
        self.num_workers = num_workers
        self.num_sems = num_sems

        self.dataloader_args = {
            "collate_fn": _tuple_collate_fn,
            "num_workers": self.num_workers,
            "worker_init_fn": worker_init_fn if self.num_workers > 0 else None,
            "persistent_workers": self.num_workers > 0,
            "pin_memory": True,
            "prefetch_factor": 16 if self.num_workers > 0 else None,
        }

        self.val_dataloader_args = {
            "collate_fn": _tuple_collate_fn,
            "pin_memory": True,
        }

        self.train_dataset: Dataset
        self.val_dataset: Dataset

        self.save_hyperparameters()

    def setup(self, stage: str):
        for num_nodes, prob, is_linear in product(self.node_nums, self.probs, (True, False)):
            shapes_dict = {f"x_{i}": torch.Size([1]) for i in range(num_nodes)}
            noise_dist_samplers = {
                f"x_{i}": UnivariateNormalNoiseModuleSampler(std_dist=td.Uniform(low=0.2, high=2.0), dim=1)
                for i in range(num_nodes)
            }
            joint_noise_module_sampler = JointNoiseModuleSampler(noise_dist_samplers)
            adjacency_dist = ErdosRenyiDAGDistribution(num_nodes=num_nodes, probs=torch.tensor(prob))

            dim = sum(shape.numel() for shape in shapes_dict.values())

            functional_relationships_sampler: Sampler[FunctionalRelationships]

            if is_linear:
                ones_matrix = torch.ones((dim, dim), dtype=torch.float32)
                functional_relationships_sampler = LinearRelationshipsSampler(
                    scale_dist=td.Uniform(
                        low=ones_matrix,
                        high=3.0 * ones_matrix,
                    ),
                    shapes_dict=shapes_dict,
                )
            else:
                num_rf = 100
                ones_matrix = torch.ones((num_rf, dim), dtype=torch.float32)
                ones = torch.ones((num_rf,), dtype=torch.float32)
                functional_relationships_sampler = RFFFunctionalRelationshipsSampler(
                    rf_dist=td.Uniform(
                        low=7.0 * ones_matrix,
                        high=10.0 * ones_matrix,
                    ),
                    coeff_dist=td.Uniform(
                        low=10.0 * ones,
                        high=20.0 * ones,
                    ),
                    shapes_dict=shapes_dict,
                )

            # store each of the sem samplers in a dictionary
            self.sems_dict[(num_nodes, prob, is_linear)] = SEMSampler(
                adjacency_dist=adjacency_dist,
                joint_noise_module_sampler=joint_noise_module_sampler,
                functional_relationships_sampler=functional_relationships_sampler,
            )

        self.train_dataset = self._get_dataset(self.train_batch_size)
        self.val_dataset = self._get_dataset(self.test_batch_size)

    def _get_dataset(self, dataset_size: int) -> Dataset:
        """Builds causal datasets given the SEM samplers.

        Args:
            dataset_size: Number of samples of the causal dataset (ie number of datasets generated).

        Returns:
            dataset object
        """
        dataset_fraction = self.num_workers if self.num_workers > 0 else 1
        factor = self.num_sems if self.num_sems > 0 else 1
        dataset = ChainDataset(
            [
                CausalDataset(
                    sampler,
                    sample_dataset_size=self.dataset_size,
                    dataset_size=dataset_size * factor * 16 // dataset_fraction,
                    num_interventions=self.num_interventions,
                    num_sems=self.num_sems,
                )
                for sampler in self.sems_dict.values()
            ]
        )

        return dataset

    def train_dataloader(self):
        return DataLoader(dataset=self.train_dataset, batch_size=self.train_batch_size, **self.dataloader_args)

    def val_dataloader(self):
        datasets = [self.train_dataset, self.val_dataset] if self.num_sems > 0 else [self.val_dataset]

        return [
            DataLoader(dataset=dataset, batch_size=self.test_batch_size, **self.val_dataloader_args)
            for dataset in datasets
        ]


def worker_init_fn(_):
    worker_info = torch.utils.data.get_worker_info()

    dataset = worker_info.dataset  # the dataset copy in this worker process
    if isinstance(dataset, ChainDataset):
        # Split chained datasets across multiple workers.
        dataset.datasets = [dataset.datasets[worker_info.id % len(dataset.datasets)]]


def _tuple_collate_fn(data: Iterable[tuple[torch.Tensor, ...]]) -> tuple[torch.Tensor, ...]:
    """Collates a list of tuples of tensors into a tuple of tensors.

    The dataloader returns batch_shape tuples of (X, y, ...), so we stack them all
    to get tensors of shapes [batch_shape, *X.shape] and [batch_shape, *y.shape] and so on.

    Args:
        data: list of tuple of tensors to collate. Assumes the dimensions of the tensors in the tuples match.

    Returns:
        collated data
    """

    def _nested_stack(x: list):
        """Stacks a tuple of tensors, returns None if an element is None, or returns lists of lists."""
        if isinstance(x[0], (torch.Tensor, TensorDict)):
            return torch.stack(x, dim=0)
        if isinstance(x[0], list):
            return list(x)
        if x[0] is None:
            return None

        raise ValueError(f"Unexpected type {type(x[0])}")

    return tuple(_nested_stack(list(x)) for x in zip(*data))



import abc

import torch
from tensordict import TensorDict

from causica.distributions.transforms import TensorToTensorDictTransform


class FunctionalRelationships(abc.ABC, torch.nn.Module):
    def __init__(self, shapes: dict[str, torch.Size]) -> None:
        """_summary_

        Args:
            shapes: Dict of node shapes (how many dimensions a node has)
                Order corresponds to the order in graph(s).
        """
        super().__init__()
        self.shapes = shapes
        # create a transform for mapping tensors to tensordicts
        self.tensor_to_td = TensorToTensorDictTransform(shapes)
        # this needs to be registered to the module, and register buffer doesn't work
        self.stacked_key_masks = torch.nn.Parameter(self.tensor_to_td.stacked_key_masks(), requires_grad=False)

    @abc.abstractmethod
    def forward(self, samples: TensorDict, graphs: torch.Tensor) -> TensorDict:
        """Calculates the predictions of the children from parents.

        Args:
            samples: dictionary of variable samples of shape sample_shape + [node shape]
            graphs: tensor of shape batch_shape + [nodes, nodes]

        Returns:
            Dictionary of torch.Tensors of shape sample_shape + batch_shape + [node shape]
        """



import torch

from causica.data_generation.samplers.functional_relationships_sampler import FunctionalRelationshipsSampler
from causica.data_generation.samplers.noise_dist_sampler import JointNoiseModuleSampler
from causica.data_generation.samplers.sampler import Sampler
from causica.distributions import AdjacencyDistribution
from causica.sem.distribution_parameters_sem import DistributionParametersSEM


class SEMSampler(Sampler[DistributionParametersSEM]):
    """Sample a SEM given adjacency, a JointNoiseModuleSampler and functional relationships distributions."""

    def __init__(
        self,
        adjacency_dist: AdjacencyDistribution,
        joint_noise_module_sampler: JointNoiseModuleSampler,
        functional_relationships_sampler: FunctionalRelationshipsSampler,
    ):
        self.adjacency_dist = adjacency_dist
        self.joint_noise_module_sampler = joint_noise_module_sampler
        self.functional_relationships_sampler = functional_relationships_sampler
        self.shapes_dict: dict[str, torch.Size] = functional_relationships_sampler.shapes_dict

    def sample(self):
        adjacency_matrix = self.adjacency_dist.sample()
        functional_relationships = self.functional_relationships_sampler.sample()
        joint_noise_module = self.joint_noise_module_sampler.sample()
        return DistributionParametersSEM(
            graph=adjacency_matrix, noise_dist=joint_noise_module, func=functional_relationships
        )



import torch
from tensordict import TensorDict

from causica.functional_relationships.functional_relationships import FunctionalRelationships
from causica.nn import DECIEmbedNN


class DECIEmbedFunctionalRelationships(FunctionalRelationships):
    """
    This is a `FunctionalRelationsips` that wraps the `DECIEmbedNN` module.
    """

    def __init__(
        self,
        shapes: dict[str, torch.Size],
        embedding_size: int,
        out_dim_g: int,
        num_layers_g: int,
        num_layers_zeta: int,
    ) -> None:
        super().__init__(shapes=shapes)

        self.nn = DECIEmbedNN(self.stacked_key_masks, embedding_size, out_dim_g, num_layers_g, num_layers_zeta)

    def forward(self, samples: TensorDict, graphs: torch.Tensor) -> TensorDict:
        return self.tensor_to_td(self.nn(self.tensor_to_td.inv(samples), graphs))



from typing import Iterable, Optional

import pandas as pd
import torch
from tensordict.tensordict import TensorDict, TensorDictBase


def convert_one_hot(
    data: TensorDict,
    one_hot_sizes: Optional[dict[str, int]] = None,
):
    """
    Args:
        data: A tensordict representing the underlying data
        one_hot_keys: A list of keys to convert to one hot encodings
    """
    # one hot encode the categorical variables
    if one_hot_sizes is None:
        return data

    new_data = data.clone()
    for key, num_classes in one_hot_sizes.items():
        cat_var = new_data[key]
        assert cat_var.shape == new_data.batch_size + (1,), "Only support 1D categorical values"
        new_data[key] = torch.nn.functional.one_hot(cat_var[..., 0].to(torch.long), num_classes=num_classes)
    return new_data


def tensordict_shapes(tds: TensorDictBase) -> dict[str, torch.Size]:
    """Return the shapes within the TensorDict without batch dimensions."""
    return {key: val.shape[len(tds.batch_size) :] for key, val in tds.items(include_nested=True, leaves_only=True)}


def tensordict_from_pandas(df: pd.DataFrame) -> TensorDict:
    """
    Create a `TensorDict` from a pandas dataframe.

    It supports pandas `MultiIndex`s of depth 2 and uses the top level index
    to determine the number of keys in the dictionary.
    """
    if isinstance(df.columns, pd.MultiIndex):
        assert df.columns.nlevels == 2, "Only support MultiIndex of depth 2."
        data = {key: df[key].to_numpy() for key in df.columns.get_level_values(0).unique()}
    else:
        # in the single index case, add a dimension to the series, so they have shape [batch, 1]
        data = {key: df[key].to_numpy()[:, None] for key in df.columns}

    return TensorDict(data, batch_size=torch.Size([len(df)]))


def identity(x):
    """
    An identity function.

    This is needed to override the default collate function for the dataloaders,
    which limit the datatypes you can return and we have `TensorDict` etc.
    """
    return x


def expand_tensordict_groups(tensordict: TensorDictBase, variable_groups: dict[str, list[str]]) -> TensorDict:
    """Expand a TensorDict with variable groups to include the individual variables.

    This only includes groups that are present in the variable groups dictionary.

    Args:
        tensordict: The TensorDict to expand
        variable_groups: A dictionary mapping group names to lists of variable names

    Returns:
        A new TensorDict with the expanded variables
    """
    assert set(variable_groups.keys()) == set(tensordict.keys()), "The variable groups must match the TensorDict"
    assert all(
        (len(v) == tensordict.get(k).shape[-1]) or (len(v) == 1) for k, v in variable_groups.items()
    ), "The variable groups must match the shape of the TensorDict"

    return TensorDict(
        {
            variable_name: tensordict.get(group_name)
            if len(variable_names) == 1
            else tensordict.get(group_name)[..., i, None]
            for group_name, variable_names in variable_groups.items()
            for i, variable_name in enumerate(variable_names)
        },
        batch_size=tensordict.batch_size,
    )


def unbind_values(td: TensorDictBase, dim: int = -1) -> TensorDictBase:
    """Return a new TensorDict with the values unbound along the given dimension.

    Negative dimensions are supported, but only if all values in the tensordict have the same number of axes. As opposed
    to `TensorDict.unbind` this supports unbinding along any dimension, not just the batch dimension.

    Args:
        td: The tensordict to unbind.
        dim: The dimension to unbind along for each value in td. Supports negative indexing, but note that when the
            values of td have different numbers of dimensions, the unbinding will act on different axis. However, if the
            values have different numbers of axes and the negative index refers to a batch dim, the batch dim cannot
            be consistently modified and an IndexError will be raised.

    Returns:
        A new nested tensordict where values unbound along dim get keys by their index.

    Examples:
        >>> td = TensorDict({"a": torch.rand(2, 3, 4), "b": torch.rand(2, 3, 4)})
        >>> unbind_values(td, dim=1)
        TensorDict({"a": TensorDict({0: torch.rand(2, 4), 1: torch.rand(2, 4), 2: torch.rand(2, 4)}), "b": ...})
    """
    num_dims = {len(tensor.shape) for tensor in td.values(include_nested=True, leaves_only=True)}
    min_dims = min(num_dims)
    max_dims = max(num_dims)

    if -min_dims > dim or dim >= min_dims:
        raise IndexError("Cannot unbind values along an axis greater than the smallest number of dims in the values")

    batch_size = td.batch_size
    if dim % min_dims < len(td.batch_size):  # If for at least some tensor, dim addresses into the batch dims
        if dim < 0 and min_dims != max_dims:
            raise IndexError(
                "Cannot unbind values along a batch dim with negative indexing when the values have different numbers "
                "of dims"
            )
        # This works because there will only be a wrap around if dim is negative, which when modifying the batch size
        # is only true when min_dims == max_dims. If it is not negative, then we know that dim < min_dims (first check).
        batch_dim = dim % min_dims
        batch_size = td.batch_size[:batch_dim] + td.batch_size[batch_dim + 1 :]

    return td.apply(
        lambda value: TensorDict({str(i): v for i, v in enumerate(value.unbind(dim=dim))}, batch_size),
        batch_size=batch_size,  # Note: Set the batch size on both the nested and outer tensordict
    )


def unbound_items(td: TensorDictBase, dim: int = -1) -> Iterable[tuple[tuple[str, ...], torch.Tensor]]:
    """Return an iterable over the (key, value) pairs for each tensor unbound along dim.

    Iterates recursively over all tensor slices in the tensordict and adds an additional key indicating the slice index
    for each element along the given dimension to unbind over.

    Example:
        >>> td = TensorDict({"a": torch.ones(2, 3, 4)}, batch_size=2)
        >>> dict(unbound_items(td, dim=0))
        {
            ("a", "0"): torch.ones(3, 4),
            ("a", "1"): torch.rand(3, 4),
        }

    Args:
        td: The tensordict to flatten.
        dim: The dimension to unbind along for each value in td.

    See Also:
        `unbind_values` for more details of how the values are unbound.
    """
    unbound = unbind_values(td, dim=dim)
    # TensorDict.items incorrectly assumes that the keys are always strings, even though they are tuples for nested
    # leaves. Since we're unbinding above, they will always be tuples below. Ignore this perceived type error.
    return unbound.items(include_nested=True, leaves_only=True)  # type: ignore



from typing import Iterable, Optional, Union

from torchmetrics import Metric, MetricCollection
from torchmetrics.wrappers import MultitaskWrapper


def create_metrics_for_variables(variables: Iterable[str], metrics: MetricCollection) -> MultitaskWrapper:
    """
    This function will create a MultiTaskWrapper obj with keys to be the variable names.

    Args:
        variables: Variables name list to calculate the metrics for.
        metrics: MetricCollection to store the metrics we want to compute.

    Returns:
        MultitaskWrapper obj with keys to be the variable names.
    """

    metrics_dict: dict[str, Union[Metric, MetricCollection]] = {
        key: metrics.clone(postfix=f".{key}") for key in variables
    }

    return MultitaskWrapper(metrics_dict)


def filter_metrics_wrapper(variable_list: Optional[list[str]], metrics_wrapper: MultitaskWrapper) -> MultitaskWrapper:
    """
    This will filter the MultiTaskWrapper obj to select only the variables in variable_list. If None, the metrics will not be filtered.

    Args:
        variable_list: List of variables to filter the metrics_dict.
        metrics_wrapper: MultitaskWrapper obj with keys to be the variable names.

    Returns:
        Filtered MultitaskWrapper obj.
    """
    if variable_list is None:
        return metrics_wrapper

    filtered_dict = {key: metrics_wrapper.task_metrics[key] for key in variable_list}

    return MultitaskWrapper(filtered_dict)



import math

import torch
from tensordict import TensorDict

from causica.functional_relationships.functional_relationships import FunctionalRelationships


class RFFFunctionalRelationships(FunctionalRelationships):
    """
    A simple random fourier feature-based functional relationship.
    The formula implemented here is:
    x_i = sqrt(2/M) * sum_{i}^{M} alpha_i sin(<w_i, pa(x_i)>)
    """

    def __init__(
        self,
        shapes: dict[str, torch.Size],
        initial_random_features: torch.Tensor,
        initial_coefficients: torch.Tensor,
        trainable: bool = False,
    ) -> None:
        """
        Args:
            shapes: Dict of node shapes (how many dimensions a variable has)
                Order corresponds to the order in graph(s).
            initial_random_features: a tensor containing the random features [num_rf, output_shape]
            initial_coefficients: a tensor containing the linear outer coefficients [num_rf,]
            trainable: whether the coefficient matrix should be learnable
        """
        super().__init__(shapes=shapes)

        assert initial_random_features.shape[0] == initial_coefficients.shape[0]
        self.num_rf = initial_random_features.shape[0]

        self.shape = self.tensor_to_td.output_shape
        assert initial_random_features.shape[1] == self.shape

        self.linear_coefficients_inner = torch.nn.Parameter(initial_random_features, requires_grad=trainable)
        self.linear_coefficients_outer = torch.nn.Parameter(initial_coefficients, requires_grad=trainable)

    def forward(self, samples: TensorDict, graphs: torch.Tensor) -> TensorDict:
        """
        Args:
            samples: Batched inputs, size batch_size_x + (concatenated_shape).
            graphs: Weighted adjacency matrix, size batch_size_g + (n, n)
        Returns:
            A Dict of tensors of shape batch_shape_x + batch_shape_g + (concatenated_shape)
        """
        return self.tensor_to_td(self.non_linear_map(self.tensor_to_td.inv(samples), graphs))

    def non_linear_map(self, samples: torch.Tensor, graph: torch.Tensor) -> torch.Tensor:
        """
        Applies the non linear function to a concatenated tensor of samples.

        Args:
            samples: tensor of shape batch_shape_x + [n_cols]
            graph: tensor of shape batch_shape_g + [n_nodes, n_nodes]
        Returns:
            tensor of shape batch_shape_x + batch_shape_g + [n_cols]
        """
        batch_shape_x = samples.shape[:-1]
        batch_shape_g = graph.shape[:-2]

        masked_graph = torch.einsum("ji,...jk,kl->...il", self.stacked_key_masks, graph, self.stacked_key_masks)

        graph_broad = masked_graph.expand(*(batch_shape_x + tuple([-1] * len(graph.shape))))
        target_shape = batch_shape_x + batch_shape_g + samples.shape[-1:]
        view_shape = batch_shape_x + (1,) * len(batch_shape_g) + samples.shape[-1:]
        # Shape batch_shape_x + batch_shape_g + (num_nodes, out_dim_g)
        samples_broad = samples.view(view_shape).expand(target_shape)

        inner_prods = torch.einsum("ij,...j,...jk->...ik", self.linear_coefficients_inner, samples_broad, graph_broad)
        return math.sqrt(2 / self.num_rf) * torch.einsum(
            "i,...ij->...j", self.linear_coefficients_outer, torch.sin(inner_prods)
        )



import abc
from typing import Generic, TypeVar

SampleType = TypeVar("SampleType")


class Sampler(Generic[SampleType], abc.ABC):
    """
    An interface of a sampler, useful for generative processes

    The interface only allows sampling one thing at a time.
    """

    @abc.abstractmethod
    def sample(self) -> SampleType:
        """Sample a sample type with given shape"""



from itertools import repeat, starmap
from typing import Optional

import numpy as np
import torch
from tensordict import TensorDict
from torch.utils.data import IterableDataset

from causica.data_generation.samplers.sem_sampler import SEMSampler
from causica.datasets.interventional_data import InterventionData
from causica.distributions.transforms import TensorToTensorDictTransform


class CausalDataset(IterableDataset):
    """A dataset that returns data from SEM samples.

    SEM samples consist of (dataset, noise, graph, interventions)

    The dataset holds samples from a SEM in a TensorDict as {node_name: [num_samples, *node_shape]}
    The noise holds samples from a SEM in a TensorDict as {node_name: [num_samples, *node_shape]}
    The graph is the causal graph of the SEM [num_nodes, num_nodes]
    The interventions are a list of InterventionData objects
    """

    def __init__(
        self,
        sem_sampler: SEMSampler,
        sample_dataset_size: int,
        dataset_size: Optional[int] = None,
        num_interventions: int = 0,
        num_intervention_samples: int = 1000,
        num_sems: int = 0,
    ):
        """
        Args:
            sem_sampler: The sampler for SEMs
            sample_dataset_size: The size of the dataset to sample from the SEM
            dataset_size: The size of this dataset, if not supplied it will be infinitely long
                It is useful to set this value to a finite size so it can be used with `ChainDataset`,
                which relies on the iterator terminating to chain the next one.
            num_interventions: The number of interventions to sample per dataset. If 0, no interventions are sampled.
            num_intervention_samples: The number of samples to use to estimate the mean.
            num_sems: The number of sems to sample the data from. If 0, each data sample is generated from a new SEM.
        """
        self.sem_sampler = sem_sampler
        self.sample_dataset_size = torch.Size([sample_dataset_size])
        self.dataset_size = dataset_size
        self.num_interventions = num_interventions
        self.num_intervention_samples = num_intervention_samples
        self.num_sems = num_sems

        self.sems = [sem_sampler.sample() for _ in range(num_sems)]
        self.td_to_tensor_transform = TensorToTensorDictTransform(self.sem_sampler.shapes_dict)
        self.cur_iter = 0

    def __iter__(self):
        """Return an iterator over samples in the dataset.

        See Also:
            CausalDataset: For a description of the format of the samples.

        Note:
            We use starmap as in `repeatfunc` https://docs.python.org/3/library/itertools.html#itertools-recipes
            This creates a generator applying `_sample `dataset_size` times, or an infinite
            generator if `dataset_size` is `None`
        """
        return starmap(self._sample, repeat(tuple(), times=self.dataset_size))

    def _sample_intervention(self, sem, tensordict_data) -> InterventionData:
        """Sample an intervention and it's sample mean from a given SEM.

        Args:
            sem: SEM to sample interventional data from.
            tensordict_data: Base data for sampling an intervention value.

        Returns:
            an intervention data object
        """
        # sample the treatment and effect variable
        treatment = np.random.choice(sem.node_names, size=1, replace=False).item()

        batch_axes = tuple(range(tensordict_data.batch_dims))
        treatment_shape = tensordict_data[treatment].shape[tensordict_data.batch_dims :]
        treatment_max = torch.amax(tensordict_data[treatment], dim=batch_axes)
        treatment_min = torch.amin(tensordict_data[treatment], dim=batch_axes)

        treatment_a = torch.rand(treatment_shape) * (treatment_max - treatment_min) + treatment_min

        intervention_a = TensorDict({treatment: treatment_a}, batch_size=torch.Size())

        intervention_a_samples = sem.do(intervention_a).sample((self.num_intervention_samples,))

        intervention_data = InterventionData(
            intervention_a_samples,
            intervention_a,
            TensorDict({}, batch_size=torch.Size()),
        )
        return intervention_data

    def _sample(
        self,
    ) -> tuple[torch.Tensor, torch.Tensor, torch.Tensor, Optional[list[InterventionData]]]:
        """Sample a new dataset and returns the data, graph and interventions if num_interventions > 0."""
        if self.num_sems > 0:
            sem = self.sems[self.cur_iter % self.num_sems]
            self.cur_iter += 1
        else:
            sem = self.sem_sampler.sample()
        noise = sem.sample_noise(self.sample_dataset_size)
        observations = sem.noise_to_sample(noise)

        interventions = None
        if self.num_interventions > 0:
            interventions = [self._sample_intervention(sem, observations) for _ in range(self.num_interventions)]

        return (observations, noise, sem.graph, interventions)



from typing import Sequence

from lightning_utilities.core.rank_zero import rank_zero_only
from mlflow.entities import Metric, Param, RunTag
from pytorch_lightning.loggers import MLFlowLogger


class BufferingMlFlowLogger(MLFlowLogger):
    """MlFlowLogger that buffers metrics on logging and flushes on finalize or when the buffer is full."""

    def __init__(self, buffer_size: int, *args, **kwargs):
        """
        Args:
            buffer_size: The maximum number of metrics to buffer before flushing
            *args: Passed to `MLFlowLogger`
            **kwargs: Passed to `MLFlowLogger`
        """
        super().__init__(*args, **kwargs)
        self._buffer_size = buffer_size
        self._buffer: list[Metric] = []
        self._original_log_batch = self.experiment.log_batch
        self.experiment.log_batch = self._buffer_log_batch_metrics(self.experiment.log_batch)

    @rank_zero_only
    def _buffer_log_batch_metrics(self, original_log_batch):
        """Returns a decorated `log_batch` that buffers metrics and flushes them when the buffer is full."""

        def log_batch(
            run_id: str,
            metrics: Sequence[Metric] = (),
            params: Sequence[Param] = (),
            tags: Sequence[RunTag] = (),
        ) -> None:
            if metrics:
                self._buffer.extend(metrics)
                if len(self._buffer) >= self._buffer_size:
                    self.flush()
            if params or tags:
                original_log_batch(run_id=run_id, params=params, tags=tags)

        return log_batch

    def get_buffer_count(self) -> int:
        """Return the current number of buffered messages."""
        return len(self._buffer)

    @rank_zero_only
    def flush(self):
        if self._buffer:
            self._original_log_batch(run_id=self.run_id, metrics=self._buffer)
            self._buffer.clear()

    @rank_zero_only
    def finalize(self, *args, **kwargs) -> None:
        self.flush()
        return super().finalize(*args, **kwargs)

    @rank_zero_only
    def __del__(self) -> None:
        self.flush()
